module ModelApi

	class ModelStruct < Struct

		class <<self
			def from_hash hash
				new *hash.values_at(*members)
			end

			unless RUBY_ENGINE == 'opal'
				attr_accessor :aspect_spec

				def from_model model
					return if model.nil?
					values = members.map do |k|
							case v = aspect_spec[k]
							when Proc
								v[model]
# XXX Пока не решено окончательно, что делать, если в спецификации указан символ. Можно использовать его
# как метод-источник, если он указан с маленькой буквы или как аспект, в котором забирать связанные модели, или тип, к которому нужно приводить полученное значение
							when Symbol
								model.send v
							else
								nil
							end
						end
					new *values
				end
			end # unless RUBY_ENGINE == 'opal'
		end

		def to_hash
			to_h.compact
		end
	end # class ModelStruct

	require 'ostruct'
	class ModelOpenStruct < OpenStruct

		class <<self
			def from_hash hash
				new hash
			end

			unless RUBY_ENGINE == 'opal'
				def from_model model
					return if model.nil?
					new model.attributes
				end
			end # unless RUBY_ENGINE == 'opal'
		end

		def to_hash
			to_h.compact
		end
	end # class ModelOpenStruct

	module HyperstackTransport

		def self.included base
			base.extend ClassMethods
		end

		module ClassMethods
=begin
При сериализации объекта он передается как хэш, у которого ключем является имя класса с префиксом !, а значением - хэш, который может использоваться для создания объекта из хэшей методом from_hash.

Если в данных находится ключ, который начинается с !, то к такому ключу добавляется дополнительный !. При десериализации мы после убирания первого ! проверяем, не продолжается строка с другого !, и считаем это сериализованным классом только если продолжается.
=end
			ESCAPE = '!'

=begin
			def serialize_params(hash)
				# receives param_name -> value pairs
				# return an object ready for to_json
				# default is just return the input hash
			end

			def deserialize_params(object)
				# recieves whatever was returned from serialize_to_server
				# (param_name => value pairs by default)
				# must return a hash of param_name => value pairs
				# by default this returns object
			end
=end

			def serialize_response(object)
				case object
				when Hash
					Hash[object.map{|k, v|
						k = "#{ESCAPE}#{k}" if k.is_a?(String) && k.start_with?(ESCAPE)
						[ k, serialize_response(v) ]
					}]
				when Array
					object.map{|v| serialize_response(v)}
				when MultiLang::StringProc # Не можем сериализовать объекты с Proc
					object.to_s
				when ModelStruct, ModelOpenStruct, MultiLang::String
					{ "!#{object.class}" => serialize_response(object.to_hash) }
				when Time, ActiveSupport::TimeWithZone
					{ "!Time" => object.to_f }
				when Date
					{ "!Time" => object.to_time.to_f }
				else
					object
				end
			end

			def deserialize_response(object)
				case object
				when Array
					object.map{|v| deserialize_response(v) }
				when Hash
					if object.size == 1 && object.first.first.start_with?(ESCAPE)
						k = object.first.first[1..-1]
						v = object.first.last
						if k.start_with?(ESCAPE) # escaped key
							{k => deserialize_response(v)}
						else
							klass = Kernel::const_get(k)
							if klass < ModelStruct || klass < ModelOpenStruct || klass <= MultiLang::String
								klass.from_hash(deserialize_response(v))
							elsif klass == Time
								klass.at v
							end
						end
					else
						Hash[object.map{|k, v| [ k, deserialize_response(v) ]}]
					end
				else
					object
				end
			end

=begin
			def serialize_dispatch(hash)
			end

			def deserialize_dispatch(object)
			end
=end

		end

	end # module HyperstackTransport

	module ActiveModelPlugin
		# Example:
		# class VitalSign < ApplicationRecord
		#   include ActiveModelPlugin
		#   model_api VitalSignApi
		# end
		def self.included base
			base.extend ClassMethods
		end
		module ClassMethods
			def model_api value = nil
				if value.nil?
					@model_api
				else
					@model_api = value
				end
			end
		end
	end # module ActiveModelPlugin

	# ClientAction может выполняться как на клиенте,
	# так и на сервере. Но используется в первую очередь
	# для организации запросов с клиента.
	class ClientOperation < Hyperstack::Operation
		class << self
			attr_accessor :aspect_struct
		end
	end

	# ServerAction выполняется только на сервере,
	# и используется для связи клиента с сервером
	#
	#	В rspec-тестах использование Hyperstack::ControllerOp вызывает ошибку 'Controler is requires'
	#	Проблема и решение описаны в doc/Hyperstack.notes "ControllerOp"
	class ServerOperation < Hyperstack::ControllerOp
		include HyperstackTransport

		class << self
			attr_accessor :run_block
			attr_accessor :model_class unless RUBY_ENGINE == 'opal'
			attr_accessor :aspect_struct
		end

		step {
			begin
				result = instance_exec params.to_h, &self.class.run_block 
				# Если есть аспект структуры, то преобразовываем данные
				if struct = self.class.aspect_struct
					if result.is_a? self.class.model_class
						result = struct.from_model result
					elsif result.is_a? Enumerable
						result = result.map{|obj| struct.from_model obj}
					end
				end
				result
			rescue Exception => ex
				Rails.logger.error "#{ex} at #{ex.backtrace.first}"
				fail ex
			end
		}
	end

=begin

При вызове метода API в него можно передавать блок, который будет исполнен сразу же после начала операции. В этом блоке можно устанавливать первоначальное значение переменной-контейнера, содержимое которое будет затем изменено после завершения операции. 

С таким же эффектом можно вынести этот блок наружу, перед обращением к методу API. Но добавление внутрь блока четче организовывает код - на начальное и на конечное состояние переменной.

			VitalSignApi.list{@server_list = ArrayStore.new}.then{|res| @server_list.concat res}

=end

=begin

Класс Transient используется для хранения изменяющихся значений. Как правило, это значение, которое имеет начальное состояние и финальное. Для получения финального значения должна быть выполнена какая-то операция, как правило, через Promise. После завершение операции в объект помещается финальное значение и объект отмечается как готовый. Если во время операции произошла ошибка, то объект отмечается как готовый, но содержащий ошибку.

Обычный Transient может сообщать о завершении, если на него подпишутся через передачу коллбэка через метод on_ready. Этот коллбэк вызывается после того, как в объект поступило значение или ошибка.
Для TransientObservable объект отмечает свое изменение через механизм Hyperstack/React - mutate&observe. Таким образом, компоненты, которые используют этот объект, обновляют рендеринг с новым значением.

Это класс используется в ModelApi для оборачиваний действия, которое возвращает не Promise, а Transient. Такой объект можно сразу же использовать в выводе. В действии можно указать стартовое значение, которое помещается в объект при начале операции; значение ошибки, которое помещается при возникновении ошибки; и значение после успешного завершения.

Стартовый хэндлер получает аргументы действия, хэндлер ошибки получает стартовое значение и ошибку, и хэндлер завершения получает стартовое значение и полученный результат.

Хэндлерам доступен текущий контекст Set-объекта, в котором хранятся args и текущее value

В action с set-объектом можно передавать блок, который должен возвращать Transient-объект. В таком случае action использует его, а не создает новый Transient. Это используется для того, чтобы использовать все время один только контейнер и во время операций только изменять его содержимое. Например, такой подход удобен при паджинации списков, получаемых с сервера.

=end
	class Transient
		include Hyperstack::State::Observable

		attr_accessor :promise
		attr_reader :value, :error
		def initialize value = nil
			@value = value
			@ready = false
			@error = nil
			self
		end

		def start value
			@ready = false
			@error = nil
			@value = value
			self
		end

		# Можно вызвать ready без изменений значения, то есть не передавая никакой аргумент.
		def ready *args
			@ready = true
			@error = nil
			if args.size == 1
				@value = args.first
			end
			@on_ready and @on_ready.each{|block| block.yield self}
			self
		end

		# Второй опциональный аргумент - новое значение. Оно может быть в том числе и nil,
		# поэтому не делаем его дефолтным nil, а проверяем, есть ли вообще оставшиеся аргументы.
		def abort error, *args
			@ready = true
			if args.size == 1
				@value = args.first
			end
			@error = error
			self
		end

		def ready?
			@ready
		end

		def error?
			!@error.nil?
		end

		def on_ready &block
			(@on_ready ||= []) << block
			self
		end
	end # class Transient

	class TransientObservable < Transient
		include Hyperstack::State::Observable

		def start value
			mutate { super }
		end

		def ready value
			mutate { super }
		end

		# Второй опциональный аргумент - новое значение. Оно может быть в том числе и nil,
		# поэтому не делаем его дефолтным nil, а проверяем, есть ли вообще оставшиеся аргументы.
		def abort error, *args
			mutate { super }
		end

		def value
			observe { super }
		end

		def error
			observe { super }
		end

		def ready?
			observe { super }
		end

		def error?
			observe { super }
		end
	end # class Transient

	class Api

		class Abort < Exception
			attr_reader :error
			def initialize error
				@error = error
			end
		end

		class << self

			# If fields nil, the default aspect will use OpenStruct with all attributes from a model
			def model model, fields = nil
				@model, @fields = model, fields
				aspect :Default, @fields
			end

			# Aspect specification:
			# nil - load into the aspect all model columns
			# Array - load listed fields
			# Hash - ключи - поля аспекта, значения - если Symbol, то поле модели, или если Proc, то результат выполнения Proc с моделью

			def aspect name, spec = nil
				case spec
				when nil
					fields = nil
				when Hash
					spec = Hash[spec.map{|k, v| [k.to_sym, v.is_a?(String) ? v.to_sym : v]}]
					fields = spec.keys
				when Array
					fields = spec.map(&:to_sym)
					spec = Hash[fields.zip fields]
				end
				if fields
					struct = ModelStruct.new *fields
					unless RUBY_ENGINE == 'opal'
						struct.aspect_spec = spec
					end
				else
					struct = Class.new(ModelOpenStruct)
				end
				const_set("Aspect#{name}", struct)
			end # def aspect

			class ActionSetup
				# TODO делать отдельно validate и add_error для клиента и сервера
				%i[param inbound outbound validate add_error].each do |macros|
					define_method macros do |*args, &block|
						(@operation_macros ||= []) << [macros, args, block]
					end
				end

				def aspect aspect
					@aspect = aspect
				end

				# Если нужно прервать операцию во время обработки в start, хэндлер
				# start должен вызвать функцию abort и передать в нее ошибку. Тогда
				# эта ошибка будет сразу же передана как Transient error.
				class SetSetup
					def initialize transient_class
						@transient_class = transient_class
					end
					# Стартовые аргументы могут изменяться в стартовом блоке
					def start *args, &block
						@start = args.size == 0 ? block : args.first
					end
					def ready *args, &block
						@ready = args.size == 0 ? block : args.first
					end
					def error *args, &block
						@error = args.size == 0 ? block : args.first
					end
				end

				def set mode = nil, &setup
					klass = case mode
									when :observable
										TransientObservable
									when nil
										Transient
									else
										raise ArgumentError.new('invalid mode')
									end
					@set = SetSetup.new klass
					@set.instance_exec(&setup) if setup
				end

				def run &block
					@run = block
				end
			end

			def action name, &setup
				# Собираем настройки в объекте action_setup
				action_setup = ActionSetup.new
				action_setup.instance_exec &setup

				aspect_struct = if aspect_options = action_setup.instance_variable_get("@aspect")
					# Если указан существующий аспект, то берем его
					# Иначе создаем по спецификации новый аспект с именем действия
					if aspect_options.is_a? Symbol
						const_get("Aspect#{aspect_options}")
					else
						aspect name.capitalize, aspect_options
					end
				else
					# Если аспект вообще не указан, данные отдаются как есть, без аспекта
					nil
				end

				# Создаем и настраиваем ControllerOp-класс, который передает данные
				# между клиентом и сервером, и который выполняет задачу на сервере
				server_operation = Class.new(ServerOperation)
				server_operation.run_block = action_setup.instance_variable_get("@run")
				server_operation.model_class = Kernel::const_get(@model) unless RUBY_ENGINE == 'opal'
				server_operation.aspect_struct = aspect_struct
				const_set("ServerOperation_#{name}", server_operation)

				# Создаем и настраиваем Operation на клиенте, который валидирует
				# параметры и потом передает задачу на сервер через ControllerOp-класс
				# Вместо с Operation создаем метод, который запускает операцию и обрабатывает
				# результат.
				# NB: ClientOperation могут работать не только в браузере, но и на стороне
				# сервера, являясь просто клиентом для ServerOperation.
				client_operation = Class.new(ClientOperation)
				client_operation.aspect_struct = aspect_struct
				client_operation.step server_operation
				const_set("ClientOperation_#{name}", client_operation)

				if operation_macros = action_setup.instance_variable_get("@operation_macros")
					operation_macros.each do |macros, args, block|
						server_operation.send macros, *args, &block
						client_operation.send macros, *args, &block
					end
				end

				if set_setup = action_setup.instance_variable_get("@set")
					api_class = self
					set_class = Class.new(Set) do
						self.api_class = api_class
						%i[start ready error].each do |handler|
							handler_config = set_setup.instance_variable_get("@#{handler}")
							next unless handler_config
							if handler_config.is_a? Proc
								define_method handler do |args|
									instance_exec(args, &handler_config)
								end
							else
								define_method handler do |args|
									handler_config
								end
							end
						end
					end
					const_set("Set_#{name}", set_class)
					transient_class = set_setup.instance_variable_get("@transient_class")

					define_singleton_method name do |*args, &block|
						set_class.new.set(client_operation, transient_class, args, &block)
					end
				else
					define_singleton_method name do |*args, &block|
						run client_operation, *args, &block
					end
				end
			end # def action

			def run operation_class, *args, &block
				block.yield if block
				operation_class.run(*args)
				.fail {|err|
					if RUBY_ENGINE == 'opal'
						msg = "#{self.name} #{name}: #{err}"
						`console.error(#{msg})`
					end
					raise err
				}
				.then {|res|
					if res.is_a?(Exception)
						if RUBY_ENGINE == 'opal'
							msg = "#{self.name} #{name}: #{res}"
							`console.error(#{msg})`
						end
						raise res
					end
					res
				}
			end

		end # class << self

		class Set
			class << self
				attr_accessor :api_class
			end
			attr_accessor :value, :args

			def abort error
				raise Abort.new error
			end

			def handle_error transient, err
				if respond_to?(:error)
					begin
						transient.abort err, error(err)
					rescue Abort => ex
						transient.abort ex.error
					end
				else
					transient.abort err
				end
			end

			# args - список, и передается в хэндлер start тоже как список
			# для того, чтобы start мог изменить аргументы.
			def set operation_class, transient_class, args, &block
				@args = args

				if block
					transient = yield self
					@value = transient.value
					if respond_to?(:start)
						begin
							@value = start(@args)
						rescue Abort => ex
							return handle_error transient, ex.error
						end
					end
				else
					@value = nil
					if respond_to?(:start)
						begin
							@value = start(@args)
						rescue Abort => ex
							return handle_error transient_class.new, ex.error
						end
					end
					transient = transient_class.new @value
				end

				transient.promise = self.class.api_class.run(operation_class, *@args)
					.then {|res|
						if respond_to?(:ready)
							begin
								transient.ready ready(res)
							rescue Abort => ex
								handle_error transient, ex.error
							end
						else
							transient.ready res
						end
					}.fail {|err|
						handle_error transient, err
					}

				transient
			end

		end # class Set

	end # class Api

end # module ModelApi

=begin

TODO
Добавлять для аспектов получение списка полей модели. Чтобы можно было в ActiveRecord запрашивать не полностью строки таблицы, а только конкретные поля.


LATER
Может быть результат выполнения блока операции передавать в саму операцию в качестве параметра result? Вдруг потом где-нибудь это пригодится.

TODO Transient
По-хорошему, нужно обернуть мьютексами проверку ready, добавление on_ready и присваивание ready,
чтобы не оказалось так, что добавлен on_ready во время присваивания ready и добавленный хэндлер 
оказался не вызван. Сейчас в однопоточном коде такой проблемы нет, но в многопоточном коде будет.
=end
