require 'lto/batch'
require 'lto/strategy'
module Batch

	# Эта модель используется только для поиска и для сериализации
	# основного рабочего класса LTO::Batch, поэтому чтобы это подчеркнуть
	# класс называется Record, а не Model.
	# Мы ничего не делегируем в вложенный внутри batch, а обращаемся
	# к нему напрямую из API и других мест, с которыми работаем.

	class Record < ApplicationRecord

		self.table_name = 'batches'
		belongs_to :user, optional: true
#		belongs_to :lab, optional: true

		validates :tid, uniqueness: { scope: [:user_id, :lab_id] }
		store :data, accessors: %i[batch_data samples_plate]

		def self.tid_generate params = {}
			scope = %i[user user_id lab lab_id]
					.select(&params.method(:key?))
					.reduce(self.all){|scope, attr| scope.where(attr => params[attr])}

			latest = scope.order(id: :DESC).pluck(:tid).first

			if latest && latest =~ /^(\D*)(\d*)$/
				"#{$1}#{$2.to_i+1}"
			else
				'0'
			end
		end

		def initialize tid: nil, strategy_class:, samples:, pool_limit:, **params
			tid ||= self.class.tid_generate params
			params[:tid] = tid
			super params
			if self.samples_plate && !samples.is_a?(Enumerable)
				number = samples.to_i
				samples = Plate::list(self.samples_plate)
				samples = samples[0...number] if number > 0
			end
			self.batch = LTO::Batch.new tid, strategy_class, samples, pool_limit: pool_limit
			batch_dirty!
		end

		before_create do
			self.state = :new
		end

		def state
			super&.then &STATES.method(:key)
		end

		def state= value
			old_value = self.state
			value = value.to_sym
			case value
			when :plan
				case old_value
				when :new   ; batch.start!
				when :ready ; batch.forward!
				else        ; return old_value # недопустимый переход
				end
				batch.plan
				batch_dirty!
			when :finished
				case old_value
				when :ready ; batch.forward!
				else        ; return old_value # недопустимый переход
				end
				batch_dirty!
			end
			super STATES[value]
		end

		def report assay_results
			batch.report assay_results
			batch_dirty!
		end

		def batch= value
			@batch = value
		end

		def batch
			@batch ||= LTO::Batch.load_init JSON.load(self.batch_data)
		end

		# После любого изменения batch-объекта нужно отмечать, что он изменен
		# через вызов этого метода.
		def batch_dirty!
			self.batch_data = JSON.dump @batch.dump_init
			nil
		end

		delegate *%i[all_samples pool_limit depth], to: :batch

		def plan
			self.batch&.plan&.map{|op| {sample: op.sample, assay: op.assay}}
		end

		delegate *%i[assays assay_results assays_waiting], to: :batch
	 	delegate *%i[samples sample_results samples_waiting samples_left], to: :batch
	 	delegate *%i[all_assays all_assay_results all_samples_left all_sample_results], to: :batch
	 	delegate *%i[planner_names all_planner_names], to: :batch

		def stages
			self.batch.traces.size
		end

		def traces
			self.batch.traces.map &:dump_init
		end

		def all_samples_count; all_samples&.count; end
		def all_samples_left_count; all_samples_left&.count; end
		def samples_waiting_count; samples_waiting&.count; end
		def assays_waiting_count; assays_waiting&.count; end
		def all_assays_count; all_assays&.count; end



	end # class Record

end # module Batch
