module MultiLang

	class String
		class << self
			def from_hash hash
				obj = allocate
				hash.each do |name, value|
					obj.instance_variable_set("@#{name}", value)
				end
				obj
			end
		end

		attr_reader :origin, :scope
		def initialize origin, scope = nil
			@origin, @scope = origin, scope
		end

		def == other
			self.class == other.class && @origin == other.origin && @scope == other.scope
		end
		alias_method :eql?, :==

		def to_s locale = nil
			MultiLang.m @origin, scope: scope, locale: locale
		end

		DEFERRED = %i[downcase upcase capitalize chomp gsub size length interpolate index empty? %]
		DEFERRED.each do |deferred|
			define_method deferred do |*args, &block|
				StringDeferred.new self, deferred, *args, &block
			end
		end

		def to_hash
			Hash[
				instance_variables.map do |name|
					[name[1..-1].to_sym, instance_variable_get(name)]
				end
			]
		end
	end

	class StringProc < String
		def initialize scope = nil, &block
			@origin, @scope = block, scope
		end
	end

	class StringDeferred < String
		attr_reader :string, :operation, :args, :kwargs, :block
		def initialize string, operation, *args, &block
			@string, @operation, @args, @block = string, operation, args, block
		end

		def == other
			self.class == other.class && @operation == other.operation && @args == other.args && @block == other.block && @string == other.string
		end
		alias_method :eql?, :==

		def to_s locale = nil
			@string.to_s(locale).send @operation, *@args, &@block
		end
		alias_method :to_str, :to_s
	end

	module StringHelpers
		# Возвращает строку перевода ключа, или, если перевод не найден, то сам ключ
		def m scope: nil, locale: nil
			MultiLang.m self, scope: scope, locale: locale
		end

		# Возвращает "обернутую" строку перевода, из которой перевод берется позже,
		# при вызове to_s, и использует локаль
		def mm scope = nil
			MultiLang.mm self, scope
		end
	end

	module ObjectHelpers
		def m string, locale: nil
			MultiLang.m string, scope: self, locale: locale
		end
		def mm string
			MultiLang.mm string, self
		end
	end

	# Возвращает строку перевода ключа, или, если перевод не найден, то сам ключ
	# Если строка на входе уже переведена, то возвращаем ее же.
	def m string, scope: nil, locale: nil
		case string
		when nil; return
		when MultiLang::String; return string.to_s
		when Proc; string = string.call
		end
		if divider = string.index('..')
			key = string[divider+2..-1]
		elsif divider = string.rindex('.')
			key = string[divider+1..-1]
		else
			key = string
		end
		return key if locale == false
		# LATER сделать объединение prefix из self и scope, и делать поиск по scope
		key
	end

	# Возвращает "обернутую" строку перевода, из которой перевод берется позже,
	# при вызове to_s, и использует локаль
	def mm string, scope = nil
		MultiLang::String.new string, scope
	end

	module_function :m, :mm

end

Object.include MultiLang::ObjectHelpers
String.include MultiLang::StringHelpers

# Строка делится на префикс (используется в scope) и на ключ, который используется
# для поиска перевода или отдается сам в качестве перевода.
# Префикс состоит из одного или нескольких имен, разделяемых точкой:
#   vital_sign.pressure.Low => Low
# Если ключ тоже должен включать точку, то он должен отделяться от значения двумя точками:
#   vital_sign..Stop. Return. => Stop. Return.
# Если ключ начинается с точки, то эта точка должна быть сразу после разделяющих двух точке:
#   vital_sign.....waiting  => ...waiting
# Если префикс отсутствует, а точки в ключе есть, то строка должна начинаться с двух точке:
#   ..Stop. Return. => Stop. Return.
#   .....waiting => ...waiting
#
# TODO По аналогии с StringProc сделать StringEval - чтобы можно было передавать с сервера на клиент

=begin

TODO

### Поиск по scope 

Если scope задана путем (model/vital_sign/sign), то поиск делается по иерархии пути, к корню.

Если scope задана классом, то поиск делается аналогично поиску констант: сначала делается поиск в лексическом scope, потом у всех ancestors (у каждого тоже в его lexical scope).


### Сохраненный scope

Нужен класс MultiLang::Scope, в объекте которого фиксируется некий scope, и через который потом можно запрашивать переводы строк, привязанные конкретно к этому scope. Такие объекты могут применяться в качестве аспектов (форматов) полей записей. Например, для поля sign_name у VitalSign можно назначить MiltiLang::Scope.new('model/vital_sign/sign'). И потом через этот scope closure получать MultiLang::String - scope.m(:temperature) -> MultiLang::String(@origin = :temperature, @scope: 'model/vital_sign/sign'). Если для Scope подается класс, то он при создании Scope ресолвится в список поиска и все созданные MultiLang::String используют уже готовый список.

Syntactic sugar - определить функцию MultiLang::Scope(scope) для создания таких объектов.


### Eval

По аналогии со StringProc делать также StringEval. Его можно использовать для передачи сериализованных объектов с сервера на клиент.

=end
