module Enumerable
	def map_hash_values &block
		Hash[[self, map(&block)].transpose]
	end
	alias :map_hash :map_hash_values

	def map_hash_keys &block
		Hash[[map(&block), self].transpose]
	end

	def map_find &block
		each { |*margs|
			res = block.call *margs
			return res if res
		}
		nil
	end

	def partition! &block
		ret = self.class.new
		delete_if { |margs|
			res = block.call(margs)
			ret << margs if res
			res
		}
		ret
	end

	def each_threaded &block
		map { |*margs|
			Thread.current.class.start(*margs, &block)
		}.each(&:join)
	end

	def map_threaded &block
		map { |*margs|
			Thread.current.class.start(*margs, &block)
		}.map(&:value)
	end

	def select_threaded &block
		selected = map { |*margs|
			Thread.current.class.start(margs) { |imargs|
				block[*imargs] ? imargs : nil
			}
		}.map(&:value).compact.map(&:first)
		self.is_a?(Hash) ? self.class[selected] : selected
	end

	def reject_threaded &block
		rejected = map { |*margs|
			Thread.current.class.start(margs) { |imargs|
				block[*imargs] ? imargs : nil
			}
		}.map(&:value).compact.map(&:first)
		self.is_a?(Hash) ? self.class[rejected] : rejected
	end

	# option: cancel_thread - block
	def find_threaded *args, &block
		if args.size > 0 && args.last.is_a?(Hash)
			options = args.pop
			cancel_action = options[:cancel_action]
		else
			cancel_action = nil
		end

		thread_to_args = map_hash_keys { |*margs|
			Thread.current.class.start(*margs, &block)
		}

		threads = thread_to_args.keys
		wait = ThreadsWait.new threads

		while !wait.empty? do
			thr = wait.next_wait
			next unless thr.value

			alive = wait.threads.select(&:alive?)

			if cancel_action
				thread_to_args.values_at(*alive).each_threaded &cancel_action
			end

			alive.each(&:kill)
			wait.all_waits rescue nil

			return thread_to_args[thr]
		end
	end

end

class Hash
	EMPTY = self.new.freeze

	def deep_compact
		Hash[map {|k,v| [k, v.is_a?(Hash) ? v.deep_compact : v] }].
			reject {|k,v| v.nil? || (v.is_a?(Hash) && v.empty?)}
	end

	def deep_compact!
		each {|k,v| v.deep_compact! if v.is_a? Hash}
		reject! {|k,v| v.nil? || (v.is_a?(Hash) && v.empty?)}
		self
	end

	def deep_store key, *keys, value
		if keys.empty?
			self[key] = value
		else
			(self[key] ||= {}).deep_store keys.shift, *keys, value
		end
	end

	def deep_fetch key, *keys, &block
		if !self.key?(key) && block
			return block.call(key, *keys)
		end
		deeper = self.fetch key
		if keys.empty?
			return deeper
		else
			deeper.deep_fetch keys.shift, *keys, &block
		end
	end

	# from Ruby on Rails
	def deep_dup
		hash = dup
		each_pair do |key, value|
			if key.frozen? && ::String === key
				hash[key] = value.deep_dup
			else
				hash.delete(key)
				hash[key.deep_dup] = value.deep_dup
			end
		end
		hash
	end

	def deep_merge(other_hash, &block)
		dup.deep_merge!(other_hash, &block)
	end

	def deep_merge!(other_hash, &block)
		merge!(other_hash) do |key, this_val, other_val|
			if this_val.is_a?(Hash) && other_val.is_a?(Hash)
				this_val.deep_merge(other_val, &block)
			elsif block_given?
				block.call(key, this_val, other_val)
			else
				other_val
			end
		end
	end

	def with_key key, &block
		block.call(self.fetch key) if self.key? key
	end

	def partition! &block
		ret = self.class.new
		delete_if { |*margs|
			res = block.call(*margs)
			ret[margs.first] = margs.last if res
			res
		}
		ret
	end

	def map_keys &block
		self.class[map {|k, v| [yield(k), v] }]
	end
	alias transform_keys map_keys

	def map_keys! &block
		replace map_keys &block
	end
	alias transform_keys! map_keys!

	def map_values &block
		self.class[map {|k, v| [k, yield(v)] }]
	end
	alias transform_values map_values

	def map_values! &block
		replace map_values &block
	end
	alias transform_values! map_values!

	def slice *keys
		keys = keys.select(&method(:key?))
		self.class[keys.zip(values_at *keys)]
	end

	def except *keys
		slice *(self.keys - keys)
	end

end

class Array
	EMPTY = self.new.freeze

	def deep_dup
		map(&:deep_dup)
	end
end

class String

	def delete_prefix prefix
		if start_with? prefix
			self[prefix.length..-1]
		else
			self
		end
	end unless instance_methods.include?(:delete_prefix)

	def delete_suffix suffix
		if end_with? suffix
			self[0...(length-suffix.length)]
		else
			self
		end
	end unless instance_methods.include?(:delete_suffix)

end

class Object
	# Динамическое создание нового класса. Мы сначала
	# создаем неинициализированный класс и сразу же
	# присваиваем его константе. И только затем его
	# инициализируем по стандартной схеме.
	# Таким образом имя нового созданного класса доступно
	# с самого начала.
	def self.define_class name, super_class, &block
		new_class = Class.allocate
		const_set(name, new_class)
		new_class.send :initialize, super_class, &block
	end

	# from Ruby on Rails
	def deep_dup
		duplicable? ? dup : self
	end

	# from Ruby on Rails
	def duplicable?
		true
	end

	def then &block
		yield self
	end unless Object.instance_methods.include?(:then)

	def nullify &block
		block.call(self) ? nil : self
	end
end

module FromRails

	# File 'activesupport/lib/active_support/inflector/methods.rb', line 92
	module_function def underscore(camel_cased_word)
		return camel_cased_word unless /[A-Z-]|::/.match?(camel_cased_word)
		word = camel_cased_word.to_s.gsub("::", "/")
		word.gsub!(/([A-Z\d]+)([A-Z][a-z])/, '\1_\2')
		word.gsub!(/([a-z\d])([A-Z])/, '\1_\2')
		word.tr!("-", "_")
		word.downcase!
		word
	end

end

module Enumerable

	def mean
		sum.to_f/size
	end

	def variance
		m = mean
		sum = reduce(0){|a, x| a +(x-m)**2 }
		sum.to_f/(size - 1)
	end

	def sdev
		Math.sqrt variance
	end

end

module InstanceVariablesEqual

	def == other
		return false unless self.class == other.class
		vars = instance_variables
		return false unless vars == other.instance_variables
		vars.all? {|var| instance_variable_get(var) == other.instance_variable_get(var) }
	end
	alias eql? ==

	def hash
		instance_variables.map(&self.method(:instance_variable_get)).hash
	end

end

module Lazy
	# extend Lazy and then
	# lazy def some_method; end
	def lazy *methods
		methods.each do |method|
			origin_method = "lazy_origin_#{method}".intern
			inst_var = "@#{method}".intern
			alias_method origin_method, method
			define_method(method) do |*args, &block|
				if instance_variable_defined?(inst_var)	
					instance_variable_get(inst_var)
				else
					result = send(origin_method, *args, &block)
					instance_variable_set(inst_var, result)
				end
			end
		end
	end

	def self.extended base
		base.class_exec do
			def lazy_clear *methods
				methods.each do |method|
					inst_var = "@#{method}".intern
					remove_instance_variable(inst_var) if instance_variable_defined?(inst_var)
				end
			end
		end
	end
end
