require 'lto/batch/stage'
module LTO
	class Batch
		class Front < Stage

=begin

В Front хранится один или несколько планировщиков этого шага и общий результат по сэмплам и тестам.
Front отвечает за маршрутизацию результатов теста в нужный планировщик и накопление этих результатов, до тех пор, пока они не будут необходимы для получения вывода из планировщика.


=end

			class AssayGroup
				include InstanceVariablesEqual

				# @param assays [Array<String>] список идентификаторов тестов
				attr_reader :planner, :results
				def initialize planner, assays 
					@planner = planner
					@results = assays.map_hash{nil}
					@submitted = false
				end

				def include? assay
					@results.include? assay
				end

				def result assay
					@results[assay]
				end

				def report assay, result
					@results[assay] = result
				end

				def filled?
					@results.values.none?(&:nil?)
				end

				def submitted?
					@submitted
				end

				def submit
					if filled?
						@planner.report @results
						@submitted = true
					else
						@submitted = false
					end
				end
			end

			include InstanceVariablesEqual

			attr_reader :planners
			def initialize planners
				@planners = planners
				@assay_groups = []
			end

			def ready?
				@plan && @assay_groups.all?(&:submitted?)
			end

			def plan
				@plan ||= begin
					@planners.map do |planner|
						curr = planner.plan
						@assay_groups << AssayGroup.new(planner, curr.map(&:assay))
						curr
					end.reduce(&:+)
				end
			end

			def forward
				raise "not ready" unless ready?
				next_planners = @planners.map(&:forward).reduce(:+)
				next_planners.empty? ? nil : self.class.new(next_planners)
			end

			# Принимает результаты тестов и накапливает их в @assay_groups
			# Если группа заполнилась полностью, то передает результаты этой группы в ее планировщик
			# Если возникает ошибка, то результаты репорта не принимаются и не передаются никуда дальше
			# @param assay_results [Hash(String, boolean)]
			def report assay_results
				unknown_assays = []
				duplicated_results = {}
				finished_planners = []

				report_queue = []	# сохраняем группы, в которые нужно будет передать результат

				assay_results.each do |assay, result|
					if group = @assay_groups.find{|group| group.include? assay}
						report_queue << [group, assay, result]
						duplicated_results[assay] = prev if prev = group.result(assay)
					else
						unknown_assays << assay
					end
				end

				ret = OperationReturn.new results: {}, warnings: {}, errors: {}
				ret.errors[:unknown_assays] = unknown_assays unless unknown_assays.empty?
				ret.warnings[:duplicated_results] = duplicated_results unless duplicated_results.empty?

				if ret.errors.empty?
					report_queue.each do |group, assay, result|
						group.report assay, result
					end

					@assay_groups.reject(&:submitted?).select(&:filled?).each do |group|
						group.submit
						finished_planners << group.planner
					end
					ret.results[:finished_planners] = finished_planners unless finished_planners.empty?
				end

				ret
			end

			def samples
				@planners.map(&:samples).reduce(&:+)
			end

			def assays
				@planners.map(&:assays).reduce(&:+)
			end

			# Финальные результаты сэмплов, только готовые сэмплы
			def sample_results
				@planners
					.select(&:ready?)
					.map(&:sample_results)
					.reduce({}, &:merge)
					.reject{|k,v| v.nil?}
			end

			# Финальные результаты тестов
			def assay_results
				@assay_groups
					.map(&:results)
					.reduce({}, &:merge)
					.reject{|k, v| v.nil?}
			end

			# Сэмплы, ожидающие текущих результатов
			def samples_waiting
				@planners
					.reject(&:ready?)
					.map(&:samples)
					.reduce([], &:+)
			end

			# Тесты, для которых ожидаются результаты
			def assays_waiting
				assays - assay_results.keys
			end

			# Сэмплы, для которых тестирование продолжается на следующем этапе
			def samples_left
				@planners
					.select(&:ready?)
					.map(&:sample_results)
					.reduce({}, &:merge)
					.select{|k,v| v.nil?}
					.keys
			end

			def planner_names
				@planners.map{|pl| pl.class.name.delete_prefix('LTO::Planner::')}.uniq
			end

			def dump_init
				{
					planners:        @planners.map{|pl| [pl.class.name.delete_prefix('LTO::Planner::'), pl.dump_init]},
					plan:            @plan&.map{|op| {sample: op.sample, assay: op.assay} },
					assay_groups:     @assay_groups.map{|tg| {
							planner: @planners.index(tg.planner),
							results: tg.results,
							submitted: tg.submitted?
						} }
				}
			end

			def self.load_init hash
				hash.transform_keys!(&:to_sym)
				planners = hash[:planners].map{|class_name, init| LTO::Planner.const_get(class_name).load_init init}
				plan = hash[:plan].map{|op|
					op.transform_keys!(&:to_sym)
					LTO::Planner::Operation.new op[:sample], op[:assay]
				} if hash[:plan]
				assay_groups = hash[:assay_groups].map{|tg|
					tg.transform_keys!(&:to_sym)
					obj = AssayGroup.allocate
					obj.instance_variable_set :@planner, planners[tg[:planner]]
					obj.instance_variable_set :@results, tg[:results]
					obj.instance_variable_set :@submitted, tg[:submitted]
					obj
				}

				obj = allocate

				obj.instance_variable_set :@planners, planners
				obj.instance_variable_set :@plan, plan
				obj.instance_variable_set :@assay_groups, assay_groups

				obj
			end



		end # class Front
	end # class Batch
end # module LTO


=begin

Для samples_forwarded можно показывать также группы, в которых находятся сэмплы, и информацию о min_positives для них.


=end
