module LTO
	module Planner

		Operation = Struct.new :sample, :assay, :volume

		class Error < RuntimeError
		end

		module Utils

			# Converts all booleans from the input to 'ones' and 'zeros'
			# Other types are left as is.
			# @param value - boolean or array or other structure
			module_function def boolean_to_integer value
				case value
				when true; 1
				when false; 0
				when Hash;
					Hash[value.map{|k, v| [send(__callee__, k), send(__callee__, v)]}]
				when Enumerable;
					value.map &method(__callee__)
				else
					value
				end
			end

			# An function opposite to boolean_to_integer
			# @param value - integer or array or other structure
			module_function def integer_to_boolean value
				case value
				when 1; true
				when 0; false
				when Hash;
					Hash[value.map{|k, v| [send(__callee__, k), send(__callee__, v)]}]
				when Enumerable;
					value.map &method(__callee__)
				else
					value
				end
			end

		end #  module Utils

		class Base

			@principal_attributes = []

			class <<self
				def principal_attributes *attrs
					@principal_attributes ||= superclass.send(__callee__).dup
					if attrs.size > 0
						@principal_attributes.concat(attrs)
						attr_reader *attrs
					end
					@principal_attributes
				end
			end

			# @param strategy [String] a strategy
			# @param samples [Integer, Array<String>] the number of samples to assay or a list of the identifiers of samples
			# @param batch [Array<Integer>] идентификатор батча
			# @param min_positives [Integer] - минимальное число позитивов, о которых можно быть уверенными
			# LATER @param assays_leased [Integer, Array<String>] the number of assays_leased (pools) available, or a list of the identifiers of the assays_leased available
			#
			# Создается объект Planner, который сразу же дает план на текущий батч и ждет его результатов. По результатам
			# создается либо финальный ответ, либо промежуточный ответ и солвер для оставшихся сэмплов, либо если коллизия
			# без промежуточного ответа, то солвер и план на следующий батч (или несколько батчей).
			# Генерируется план сборки тестовых пулов из сэмплов и схема определения по результатам тестов какие сэмплы позитивные.
			#
			attr_reader :strategy, :samples
			principal_attributes :batch, :min_positives
			def initialize strategy, samples, batch: [], min_positives: 0, **kwargs
				samples = (1..samples).to_a.map{|s| "S#{s}"} unless samples.is_a?(Enumerable)
				@strategy, @samples, @batch, @min_positives = strategy, samples, batch, min_positives
			end

			# Список тестов, запланированных к использованию в текущем плане
			attr_reader :assays

			# Возвращает имена тестов.
			# Может быть переопределена для того, чтобы резервировать тесты в выделенном списке тестов.
			def request_assays count
				(1..count).to_a.map{|s| "T#{batch.join('-')}/#{s}"}
			end

			# Создание плана операций
			# Должен наполнить значение `@assay_samples`
			# @abstract
			protected def build_plan
				raise NotImplementedError
			end
			attr_reader :assay_samples
			# В @assay_samples для каждого теста указываются номера сэмплов, попадающих в группу этого сэмпла

			# План текущего батча, в виде списка операций какой сэмпл в какой тест добавлять.
			# Этот внешнее представление плана, которое строится по уже подготовленному внутреннему плану.
			# Если план возвращает пустой список (@assay_samples пуст), то клиенту следует сразу же запросить
			# следующие планировщики через next_forward.
			def plan
				operations = []
				@assay_samples.each_with_index do |samples, i|
					assay = @assays[i]
					samples.each do |j|
						operations << Operation.new(@samples[j], assay)
					end
				end
				operations
			end

			# Принять результаты тестов и сохранить заключение в планировщике.
			# Заключение состоит из результатов сэмплов и краткого резюме шага.
			# Результат сэмпла - это либо финальный результат positive (true) or negative (false), или неопределенный вариант nil.
			# Краткое резюме может включать
			#   :hit, если получены финальные результаты и больше не требуется продолжения
			#   :miss, если требуется дальнейшее продолжение
			#   :error, если результат тестов показывает ошибочность текущего или предыдущего шага,
			#   :futile, если шаг не содержит ошибки, но не способен дать дополнительную информацию
			# @abstract
			# @param results [Hash{String => Number}] - результат тестов в виде идентификатора теста и его значения. Значение может быть как `0` или `1`, так и число, например, скорость роста для real-time PCR.
			# @return [Hash{String => Number}, nil], [Planner, nil], [Symbol, nil] - возвращается результат по сэмплам, планировщик следующего батча, и оценка общих результатов батча
			# Результаты - false, true, или nil. nil может означать, что этот тест по какой-то причине не удался
			def report results
				requested_assays = @assay_samples.each_index.map{|i| @assays[i]}
				raise Error.new("missing results for assays #{requested_assays-results.keys}") if results.keys != requested_assays
				assay_output = Array.new(@assays.size)
				reported_assays = []
				missing_assays = []
				results.each do |assay, result|
					if i = @assays.index(assay)
						assay_output[i] = result
						reported_assays << assay
					else
						missing_assays << assay
					end
				end
				raise Error.new("missing assays #{missing_assays.join(', ')}") unless missing_assays.empty?

				@assay_output = assay_output
				@assay_results = results

				results = self.infer
				@sample_results = Hash[results.each_with_index.map{|value, sample| [@samples[sample], value]}]
			end

			attr_reader :assay_output
			attr_reader :assay_results
			attr_reader :sample_results

			# Обработка принятых результатов тестов
			protected def infer
				raise NotImplementedError
			end


			# Перейти на следующий этап проверки и получить новые планировщики
			# @abstract
			def forward
				@forward ||= next_planners
			end

			# Если планировщик принял результаты тестов и подготовил результаты сэмплов
			def ready?
				!@sample_results.nil?
			end

			# @abstract
			protected def next_planners
				raise NotImplementedError
			end

			# Когда планировщик завершил тестирование сэмплов своими собственными средствами,
			# включая внутренние планировщики, он передает оставшиеся сэмплы на планирование
			# в Strategy. Strategy может реализовывать какие-то простые линейные стратегии
			# передачи по планировщикам, или принимать более сложные решения по числу сэмплов,
			# min_positives, или каким-то другим данным.
			def strategy_planners samples, **kwargs
				@strategy.planners self.class, samples, **kwargs
			end

			@init_attributes = []

			class <<self
				def init_attributes *attrs
					@init_attributes ||= superclass.send(__callee__).dup
					@init_attributes.concat(attrs) if attrs.size > 0
					@init_attributes
				end
			end
			init_attributes :@assay_output, :@assay_results, :@sample_results

			def dump_init
				ret = {
					samples: @samples,
				}
				ret[:strategy] = [@strategy.class.name, @strategy.dump_init] if @strategy
				self.class.principal_attributes.each do |var|
					inst_var = "@#{var}".to_sym
					ret[var] = instance_variable_get(inst_var) if instance_variable_defined?(inst_var)
				end
				self.class.init_attributes.each do |var|
					ret[var] = instance_variable_get(var) if instance_variable_defined?(var)
				end
				ret
			end

			def self.load_init hash
				hash.transform_keys!(&:to_sym)
				# strategy must be provided as a normal ruby object
				strategy_spec, samples = hash.values_at :strategy, :samples
				strategy = strategy_spec ? Kernel.const_get(strategy_spec.first).load_init(strategy_spec.last) : nil
				obj = new strategy, samples, **hash.except(:strategy, :samples, *init_attributes)
				hash.slice(*init_attributes).each do |var, value|
					obj.instance_variable_set var, value
				end
				obj
			end

		end # class Base

	end # module Planner
end # module LTO


=begin

LATER
Заранее давать список доступных идентификаторов тестов, assays_leased. И по мере создания новых батчей отмечать какие тесты уже заняты.

=end
