require 'lto/planner'
require 'lto/planner/direct'
require 'ruby_helper'
module LTO
	module Planner

		class MultipoolEnumerator < Base

			module BitmapOps
				# Bitmap - is an Array of 'true' or 'false' values.

				module_function def increment array, spare = 0
					result = array.dup
					(result.size-1).downto(0).each do |i|
						if result[i] == false
							result[i] = true
							return result
						end
						result[i] = false
					end
					# overflow
					nil
				end

				# To ensure 0! == 1 reduce always receives '1' as a start argument
				module_function def enumerate_cardinality digits, min_ones, max_ones
					min_ones, max_ones = max_ones, min_ones if min_ones > max_ones
					return 0 if digits == 0
					digits_fac = (1..digits).reduce(1, :*)
					(min_ones..max_ones).map { |i|
						digits_fac / (
							(1..i).reduce(1, :*) * (1..(digits-i)).reduce(1, :*)
						)
					}.reduce(0, :+)
				end

				# Распределение номеров по уникальным битовым маскам.
				# Биты распределяются равномерно по всем разрядом. Плотность
				# масок растет от начала к концу.
				module_function def enumerate count, dense = 0, spare = 0
					digits = dense + spare
					dense_max = dense
					loop do
						cardinality = enumerate_cardinality digits, dense, dense_max
						break if cardinality >= count
						digits += 1
						dense_max += 1
					end

					result = []
					(dense..dense_max).each do |ones|
						result.concat permutate_swap(ones, digits - ones)
					end

					result[0...count]
				end

				module_function def permutate_swap ones, zeroes
					array = [true] * ones + [false] * zeroes
					result = [array]
					queue = [[0, (array.size-1), array]]
					while (start, final, curr = queue.shift) do
						# Если удается сделать перестановку, то добавляем полученное в result
						# и добавляем в стек два варианта для дальнейших перестановок
						(start..final-1).each do |i|
							if curr[i] == true && curr[i+1] == false
								swap = curr.dup
								swap[i], swap[i+1] = swap[i+1], swap[i]
								result << swap
								queue << [i+1, final, swap]
								queue << [0, i, swap]
							end
						end
					end
					result
				end


				# Проверяем, может ли компонент входить в результат, который
				# был получин из компонентов по or. Для этого все единицы
				# компонента должны присутствовать в результате. Нули в компоненте
				# не важны. То есть, result & component == component
				module_function def match_or_result result, component
					pairs = result.zip(component)
					component == pairs.map{|a, b| a & b}
				end

				module_function def or vector_a, vector_b
					vector_a.zip(vector_b).map{|a, b| a | b}
				end

				module_function def and vector_a, vector_b
					vector_a.zip(vector_b).map{|a, b| a & b}
				end

				module_function def xor vector_a, vector_b
					vector_a.zip(vector_b).map{|a, b| a ^ b}
				end

				# @param result [Bitmap] вектор результата, который должен образоваться через OR из комбинации
				# @param components [Array<Integer>] список номером строк, из которых находится комбинация
				module_function def find_min_or_combination result, components
					_find_min_or_combination result, Array.new(result.size, false), components, components.size
				end

				# @returns Bitmap
				module_function def _find_min_or_combination result, accumulator, components, limit
					return if limit == 0
					found = nil
					components.each_with_index do |component, i|
						curr = self.or accumulator, component
						next if curr == accumulator
						return [i] if curr == result
						tail = _find_min_or_combination result, curr, components[(i+1)..-1], (found&.size || limit) - 1
						if tail
							found = [i] + tail.map{|x| i+1+x}
						end
					end
					found
				end

				# Находим все возможные комбинации компонентов, которые образуют result через OR.
				# @param result [Bitmap] вектор результата, который должен образоваться через or из комбинации
				# @param combinations [Array<Array<Integer>>] список комбинаций - список списков номеров строк, из которых находится комбинация
				module_function def find_min_or_combinations result, components
					_find_min_or_combinations result, Array.new(result.size, false), components, components.size
				end

				# @returns Bitmap
				module_function def _find_min_or_combinations result, accumulator, components, limit
					return Array::EMPTY if limit == 0 || components.empty?
					found = []
					min_length = limit

					components.each_with_index do |component, i|
						curr = self.or accumulator, component
						if curr == accumulator
							next
						elsif curr == result
							min_length = 1
							found.reject!{|combination| combination.size > min_length}
							found << [i]
						else
							combinations = _find_min_or_combinations result, curr, components[(i+1)..-1], [min_length, limit-1].min
							if combinations.size > 0
								min_length = combinations.first.size + 1
								found.reject!{|combination| combination.size > min_length}
								combinations.each do |tail|
									found << ([i] + tail.map{|x| i+1+x})
								end
							end
						end
					end

					found
				end

				# Подсчитывает число true элементов для каждого столбца
				module_function def matrix_sum matrix
					return Array::EMPTY if matrix.empty?
					(0...matrix.first.size).map do |i|
						matrix.map{|row| row[i] ? 1 : 0}.sum
					end
				end

			end # module BitmapOps

			# @param dense [Integer] минимальное число бит, которые должны присутствовать
			# @param spare [Integer] число бит, которые должны остаться нулевыми - увеличивает число тестов, но уменьшает число коллизий
			principal_attributes :dense, :spare
			def initialize strategy, samples, dense: 1, spare: 1, **kwargs
				@dense, @spare = dense, spare
				super strategy, samples, **kwargs
				build_plan
			end

			def build_plan
				@matrix = BitmapOps.enumerate(@samples.size, @dense, @spare)

				groups_count = @matrix.first.size

				# Проверяем, нет ли у нас переполнения пула
				@max_pool = (0...groups_count).map{|group| @matrix.count{|row| row[group]}}.max
				raise "too many samples #{@max_pool} for pool_limit #{@strategy.pool_limit}" if @max_pool > @strategy.pool_limit

				@assays = request_assays groups_count

				@assay_samples = @assays.each_index.map do |i|
						# Берем из матрицы те строчки, в которых этот элемент равен единице
						samples = []
						@matrix.each_with_index do |v, j|
							samples << j if v[i] == true
						end
						samples
					end

#				puts "max_pool: #{@max_pool}, groups_count: #{groups_count}"
			end
			attr_reader :matrix
			init_attributes :@matrix

			def infer
				return Array.new(@samples.size, false) if @assay_output.none?

				# Проверяем, какие сэмплы пересекаются с комбинацией assay_results.
				# Если у какого-то сэмпла получен негативный результат на тест, то это сразу же false
				# Остальные сэмплы, для которых все групповые тесты дали позитив, идут на дальнейшую проверку.
				results = Array.new @samples.size
				matched_samples = []
				@matrix.each_with_index.map do |row, sample|
#					if !BitmapOps.match_or_result @assay_output, row
					if @assay_output.each_with_index.any?{|result, assay| row[assay] & !result }
						results[sample] = false
					else
						matched_samples << sample
					end
				end


				@shadow_mask = Array.new(@matrix.first.size, false)

				# Считаем сколько проверяемых сэмплов включает каждый тест. Если только один сэмпл,
				# то это значит, что позитив вызван именно этим сэмплом и отмечаем этот сэмпл как позитив.
				# Добавляем биты этого сэмпла в shadow_mask, т.к. они затеняют тесты для других сэмплов
				# и делают их неинформативными.
				matched_matrix = matched_samples.map{|i| @matrix[i]}
				BitmapOps.matrix_sum(matched_matrix).each_with_index do |count, assay|
					next unless count == 1
					sample = matched_samples.find{|sample| @matrix[sample][assay]}
					results[sample] = true
					@shadow_mask = BitmapOps.or(@shadow_mask, @matrix[sample])
				end

				if @min_positives > 0 && results.count{|result| result != false} < @min_positives
					@error = "too few matched candidates for min_positives #{@min_positives}"
					return
				end

				# На следующий этап проверки идут сэмплы, которые совпали полностью с позитивными тестами,
				# но имеются также другие сэмплы с таким же совпадением, поэтому нет однозначного позитивного заключения.
				@left_samples = results.each_with_index.map{|result, sample| result.nil? ? sample : nil}.compact

				results
			end
			attr_reader :left_samples
			init_attributes :@shadow_mask, :@left_samples

			def next_planners
				descendants = []
				@rich_samples = []
				@poor_samples = []
				return descendants unless @left_samples && @left_samples.size > 0

				# Разделяем затеняемые и незатеняемые сэмплы по маске:
				# 	- для незатеняемых (rich) продолжается уточнение
				# 	- затеняемые (poor) идут в дефолтное планирование

				rich_mask = @shadow_mask.map(&:!)

				@left_samples.each do |sample|
					if BitmapOps.and(@matrix[sample], rich_mask).any?
						@rich_samples << sample
					else
						@poor_samples << sample
					end
				end

				if @rich_samples.size > 0
					# Сокращаем матрицу: берем оставшиеся сэмплы только с теми колонками, которые являются
					# значащими для них
					rich_assays = (0...@assay_output.size).map do |assay|
						@rich_samples.any?{|sample| @matrix[sample][assay] } ? assay : nil
					end.compact

					samples = @rich_samples.map{|i| @samples[i]}

					if rich_assays.size == 1
						# Все сэмплы оказались в одной группе и нет никакой информации о них, кроме того что
						# как-минимум один из них позитив. Поэтому вместо Clarification используется Bisector.
						descendants.concat strategy_planners(samples, batch: @batch+[0], min_positives: 1)
					else
						@rich_matrix = @rich_samples.map do |sample|
							rich_assays.map{|assay| @matrix[sample][assay] }
						end
						descendants << Clarificator.new(self, samples, batch: @batch+[0], matrix: @rich_matrix)
					end
				end

				if @poor_samples.size > 0
					# we have no idea whether the samples contain a positive, so min_positive is not set
					samples = @poor_samples.map{|i| @samples[i]}
					descendants.concat strategy_planners(samples, batch: @batch+[0])
				end

				descendants
			end
			attr_reader :rich_matrix
			attr_reader :rich_samples
			attr_reader :poor_samples


			# Планировщик, уточняющий тесты после первого прохода Enumerator
			#	Проверяем сэмпл с самым большим число единичных битов и самой низкой вероятностью быть позитивом:
			#		если он негатив, то значит для остальных можно искать уникальные биты и отметить их как позитивные,
			#			и потом продолжить оставшуюся матрицу в этом же планировщике
			#		если он позитив, то по нему дается заключение сразу же, а оставшиеся сэмплы делятся по двум планировщикам
			#			в первый планировщик Clarificator попадают сэмплы, у которых есть значимые биты, которые
			#				не пересекаются с битами найденного позитивв
			#			во второй планировщик Direct (или Bisector) попадают оставшиеся сэмплы, то есть те, у которых все значимые
			#				биты покрываются битами найденного позитива
			class Clarificator < Base

				principal_attributes :matrix
				def initialize strategy, samples, batch: [], matrix: , **kwargs
					@matrix = matrix
					super strategy, samples, batch: batch, **kwargs
					build_plan
				end


				def build_plan
					# Находим минимальную комбинацию сэмплов, участвующих в наибольшем числе тестов.
					#
					# Ищем наименьшее число покрывающих комбинаций и берем все эти комбинации.
					bit_count_order = @matrix.map{|row| row.count{|x| x} }.each_with_index.sort.map(&:last)
					sorted_matrix = bit_count_order.map{|sample| @matrix[sample]}
					combination_sorted = BitmapOps.find_min_or_combination(Array.new(@matrix.first.size, true), sorted_matrix)

					raise Error.new("matrix doesn't cover all assays") if combination_sorted.empty?

					max_combination = combination_sorted.map{|order| bit_count_order.index order}


					@assays = request_assays max_combination.size

					# В тесты идут по отдельности каждый сэмпл из комбинации
					@assay_samples = max_combination.map{|sample| [sample]}

					# Все остальные сэмплы ждут результата
					@wait_samples = (0...@samples.size).to_a - max_combination
				end
				init_attributes :@wait_samples


				def infer
					results = Array.new @samples.size, nil

					@shadow_mask = Array.new(@matrix.first.size, false)
					left_matrix = @matrix.deep_dup # матрица для следующего уточнения

					@assay_output.each_with_index do |result, assay|
						sample = @assay_samples[assay].first # Только один сэмпл в каждом тесте
						if result
							# Если позитивный тест, то отмечаем сэмпл позитивом и отмечаем
							# его тесты в матрице как затеняемые, то есть для остальных сэмплов эти
							# тесты из матрицы стали неинформативными.
							results[sample] = true
							@shadow_mask = BitmapOps.or(@shadow_mask, @matrix[sample])
						else
							# Если тест негативный, отмечаем сэмп негативом и очищаем строку
							# сэмпла в матрице для следующего уточнения.
							results[sample] = false
							left_matrix[sample] = Array.new(@matrix.first.size, false)
						end
					end

					# Ищем в матрицы уточнения единичные сэмплы и отмечаем как позитивные,
					# при этом тесты, в которых находятся такие сэмплы затеняются.

					BitmapOps.matrix_sum(left_matrix).each_with_index do |count, assay|
						next unless count == 1
						sample = left_matrix.find_index{|row| row[assay]}
						results[sample] = true
						@shadow_mask = BitmapOps.or(@shadow_mask, left_matrix[sample])
					end

					@left_samples = results.each_with_index.map{|result, sample| result.nil? ? sample : nil}.compact

					results
				end
				init_attributes :@shadow_mask, :@left_samples


				def next_planners
					descendants = []

					# Разделяем затеняемые и незатеняемые сэмплы по маске:
					# 	- для незатеняемых (rich) продолжается уточнение
					# 	- затеняемые (poor) идут в дефолтное планирование

					rich_mask = @shadow_mask.map(&:!)

					@rich_samples = []
					@poor_samples = []
					@left_samples.each do |sample|
						if BitmapOps.and(@matrix[sample], rich_mask).any?
							@rich_samples << sample
						else
							@poor_samples << sample
						end
					end

					if @rich_samples.size > 0
						# Сокращаем матрицу: берем оставшиеся сэмплы и колонки тестов, для которых отмечены
						# у rich сэмплов, и которые не затенены найденными позитивными сэмплами.
						rich_assays_mark = (0...@matrix.first.size).map do |assay|
									@rich_samples.any?{|sample| @matrix[sample][assay] }
								end

						rich_assays = BitmapOps.and(rich_assays_mark, rich_mask).each_with_index.map{|is_set, assay| is_set ? assay : nil}.compact

						samples = @rich_samples.map{|i| @samples[i]}

						if rich_assays.size == 1
							# Все сэмплы оказались в одной группе и как-минимум один из них позитив.
							descendants.concat strategy_planners(samples, batch: @batch+[0], min_positives: 1)
						else
							@rich_matrix = @rich_samples.map do |sample|
								rich_assays.map{|assay| @matrix[sample][assay] }
							end
							descendants << self.class.new(@strategy, samples, batch: @batch+[0], matrix: @rich_matrix)
						end
					end

					if @poor_samples.size > 0
						# we have no idea whether the samples contain a positive, so min_positive is not set
						samples = @poor_samples.map{|i| @samples[i]}
						descendants.concat strategy_planners(samples, batch: @batch+[0])
					end

					descendants
				end
				attr_reader :rich_matrix
				attr_reader :rich_samples
				attr_reader :poor_samples

				def strategy_planners samples, **kwargs
					@strategy.planners LTO::Planner::MultipoolEnumerator, samples, **kwargs
				end

			end


		end # class MultipoolEnumerator
	end # module Planner
end # module LTO


=begin

NB

У данной реализации Clarificator уточнение идет по минимальной комбинации сэмплов с наибольшим числом значащих бит.

Если rich_samples присутствуют, то их всегда больше одного. Если бы rich_samples был только один, то это был бы однозначный позитив.


LATER

если сэмплы не плотно забивают всю матрицу (т.е. их число заметно меньше, чем n**2-2), стоит увеличить spare

в infer если matched.size < min_positives - значит произошла ошибка


TODO
Если сэмплы не помещаются в текущий пул, их нужно оставить в left и передать в следующий планировщик


=end
